unit necklace;

interface

procedure init;
procedure create (from : longint; operation : char; on_left : boolean; param : longint);
function pearl (neck_id : longint; on_left : boolean) : longint;

implementation

type pnode = ^node;
     node = record
              v, hl : longint;
	      l, r : pnode;
	    end;
     neck = pnode;

function dpth(n:pnode):longint;
begin
  if n = nil then dpth := 0 else dpth := n^.hl;
end;

function max(a,b: longint): longint;
begin
  if a>b then max:=a else max:=b;
end;

function anode(v:longint; l, r:pnode; left:boolean) : pnode;
var n:pnode;
begin
  new(n);

  n^.v := v;
  n^.hl := max(dpth(l), dpth(r)) + 1;
  if left then
    begin
      n^.l := l;
      n^.r := r;
    end
  else
    begin
      n^.l := r;
      n^.r := l;
    end;

  anode := n;
end;

function size(n:pnode):longint;
begin
if n = nil then size := 0 else size := 1 + size(n^.l) + size(n^.r);
end;

function lson(n:pnode; left:boolean):pnode;
begin
  if left then lson := n^.l else lson := n^.r;
end;

function rson(n:pnode; left:boolean):pnode;
begin
  if left then rson := n^.r else rson := n^.l;
end;

function ins(n:pnode; v : longint; left:boolean):pnode;
var r, nl, a, b, nr, bl, br, drl : pnode;
    dl, dr : longint;
begin
  if n = nil then
    begin
      ins := anode (v, nil, nil, left);
      exit;
    end;

   nl := ins (lson(n, left), v, left);
   r := rson (n, left);
   dl := dpth (nl);
   dr := dpth (r);
   if dr + 1 >= dl then
     begin
       ins := anode (n^.v, nl, r, left);
       exit;
     end;

   assert (dl = dr + 2);
   a := lson (nl, left);
   b := rson (nl, left);

   if dpth(a) = dr + 1 then
     begin
       nr := anode(n^.v,b,r,left);
       ins := anode (nl^.v,a,nr,left);
       exit;
     end;

   assert (dpth (b) = dr + 1);
   bl := lson (b, left);
   br := rson (b, left);

   nr := anode (n^.v, br, r, left);
   drl := anode (nl^.v, a, bl, left);
   ins := anode (b^.v, drl, nr, left);
end;

function del(n:pnode; left:boolean):pnode;
var l, r, nl, b, c, x, y, bl, br : pnode;
    dl, dr : longint;
begin
  l := lson(n, left);
  r := rson(n, left);
  if l = nil then
    begin
      del := r;
      exit;
    end;

  nl := del(l,left);
  dr := dpth (r);
  dl := dpth (nl);

  if dl + 1 >= dr then
    begin
      del := anode(n^.v, nl, r, left);
      exit;
    end;

  assert (dl + 2 = dr);
  b := lson(r, left);
  c := rson(r, left);

  if dpth (c) = dl + 1 then
    begin
      x := anode (n^.v, nl, b,left);
      del := anode (r^.v, x, c,left);
      exit;
    end;

  assert (dpth (b) = dl + 1);
  bl := lson (b, left);
  br := rson (b, left);

  x:=anode (n^.v, nl, bl,left);
  y := anode (r^.v, br, c,left);
  del := anode (b^.v, x, y,left);
end;

procedure pushF (n : neck; v : longint; var nw : neck);
begin
  nw := ins (n, v, true);
  {assert (size(nw) = size(n) + 1);}
end;

procedure pushR (n : neck; v : longint; var nw : neck);
begin
  nw := ins (n, v, false);
  {assert (size(nw) = size(n) + 1);}
end;

procedure popF (n : neck; var nw : neck);
begin
  nw := del (n, true);
  {assert (size(nw) = size(n) - 1);}
end;

procedure popR (n : neck; var nw : neck);
begin
  nw := del (n, false);
  {assert (size(nw) = size(n) - 1);}
end;

function topF (n : neck) : longint;
begin
  while n^.l <> nil do n := n^.l;
  topF := n^.v;
end;

function topR (n : neck) : longint;
begin
  while n^.r <> nil do n := n^.r;
  topR := n^.v;
end;

procedure empty_neck (var nck: neck);
begin
 nck := nil;
end;

var necks:array[0 .. 1000000] of neck;
    nnecks : longint;

procedure init;
begin
 nnecks := 1;
 empty_neck (necks[0]);
end;

procedure create (from : longint; operation : char; on_left : boolean; param : longint);
begin
  if operation = 'A' then
    begin
      if on_left then pushF (necks[from], param, necks[nnecks])
                 else pushR (necks[from], param, necks[nnecks]);
    end
  else
    begin
      if on_left then popF (necks[from], necks[nnecks])
                 else popR (necks[from], necks[nnecks]);
    end;
  inc (nnecks);
end;

function pearl (neck_id : longint; on_left : boolean) : longint;
begin
  if on_left then
    pearl := longint(topF (necks[neck_id]))
  else
    pearl := longint(topR (necks[neck_id]));
end;

procedure dump_necklace(nck : neck);
begin
  if nck <> nil then
    begin
      dump_necklace(nck^.l);
      write (' ', nck^.v);
      dump_necklace(nck^.r);
    end;
end;

procedure dump_necklaces;
var i : longint;
begin
  for i := 0 to nnecks - 1 do
    begin
    writeln ('Necklace ', i);
    dump_necklace (necks[i]);
    end;
end;

end.

