#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#define MAXPEARLS 2000000
#define MAXNECKLACES 2000000

//pearls form stacks as head:tail lists
struct PEARL
{
	unsigned color;		//pearl color 
	struct PEARL *next;	//tail of the list
	struct PEARL *suspended;//if not NULL, this pearl is suspended
				//suspended stack consists of cdirect pearls 
				//after *next and creverse pearls reversed 
				//from stack with head cskip nexts from *suspended
	unsigned cdirect;	//if suspended -- number of used items 
				//from old stack
	unsigned creverse;	//if suspended -- number of items to 
				//reverse from former oposite stack
	unsigned cskip;		//if suspended -- number of nexts to skip from 
				//*suspended
};
// even suspended pearl contains valid color, but next is not real tail

struct NECKLACE {
	unsigned llen,rlen;	//lengths
	struct PEARL *l,*r;	//stacks
	unsigned num;		//number
};

// global pearls array
struct PEARL pearls[MAXPEARLS];
int cpearls;

// global necklaces array
struct NECKLACE necklaces[MAXNECKLACES];
int cnecklaces;

// debug operations info
int cnecklace_appends;
int cnecklace_removals;
int cresume_ops;

//global temp variable used by necklace2str() and necklace_status()
char glob_str[10*300000];

//forward decl
void pearl_resume(struct PEARL *p);

//get new necklace
struct NECKLACE* new_necklace()
{
	assert(cnecklaces<MAXNECKLACES);
	cnecklaces++;
	necklaces[cnecklaces-1].num=cnecklaces-1;
	return &(necklaces[cnecklaces-1]);
}

//get new pearl
struct PEARL *new_pearl()
{
	assert(cpearls<MAXPEARLS);
	cpearls++;
	pearls[cpearls-1].suspended=NULL;
	return &(pearls[cpearls-1]);
}
	
// if llen and rlen differ much, balance them by 
// cutting the longer stack and creating new suspended 
// stack head instead the other stack
inline void necklace_rebalance(struct NECKLACE *n)
{
	int len=n->llen+n->rlen;
	int llen=len/2;
	int rlen=len-llen;
	if((len>=4)&&(n->llen>len/4)&&(n->rlen>len/4))
		return;
	if(len<=1) return;
	if(len==2)
	{
		// fix cases 0:2, 2:0
		if(n->llen==0)
		{
			struct PEARL *p=new_pearl();
			p->next=NULL;
			pearl_resume(n->r);
			p->color=n->r->next->color;
			n->l=p;
		}
		if(n->rlen==0)
		{
			struct PEARL *p=new_pearl();
			p->next=NULL;
			pearl_resume(n->l);
			p->color=n->l->next->color;
			n->r=p;
		}
		n->llen=1;
		n->rlen=1;
		return;
	}
	if(len==3)
	{
		// fix cases 0:3, 3:0
		if(n->llen==0)
		{
			struct PEARL *p=new_pearl();
			p->next=NULL;
			pearl_resume(n->r);
			pearl_resume(n->r->next);
			p->color=n->r->next->next->color;
			n->l=p;
			n->llen=1;
			n->rlen=2;
		}
		if(n->rlen==0)
		{
			struct PEARL *p=new_pearl();
			p->next=NULL;
			pearl_resume(n->l);
			pearl_resume(n->l->next);
			p->color=n->l->next->next->color;
			n->r=p;
			n->rlen=1;
			n->llen=2;
		}
		return;
	}
	assert(n->llen>0);
	assert(n->rlen>0);
	if(n->llen>n->rlen)
	{//shorten left one, suspend right
		struct PEARL *p=new_pearl();
		pearl_resume(n->r);
		p->color=n->r->color;
		p->next=n->r->next;
		p->cdirect=n->rlen-1;
		p->cskip=llen;
		p->creverse=n->llen-llen;
		p->suspended=n->l;
		n->r=p;
	}
	else
	{//shorten right one, suspend left
		struct PEARL *p=new_pearl();
		pearl_resume(n->l);
		p->color=n->l->color;
		p->next=n->l->next;
		p->cdirect=n->llen-1;
		p->cskip=rlen;
		p->creverse=n->rlen-rlen;
		p->suspended=n->r;
		n->l=p;
	}
	n->llen=llen;
	n->rlen=rlen;
}

// if this pearl is suspended, make it regular and create next 
// suspended pearl or reverse what remains
// !!!! BEFORE LOOKING AT PEARL'S next, RESUME IT !!!!
void pearl_resume(struct PEARL *p)
{
	unsigned cnt;
	struct PEARL *tail, *stack;
	if(!p->suspended) return;
	if(p->cdirect>0)
	{//create new suspended, resume this one
		struct PEARL *n=new_pearl();
		assert(p->next!=NULL);
		pearl_resume(p->next);
		n->suspended=p->suspended;
		n->color=p->next->color;
		n->next=p->next->next;
		n->cdirect=p->cdirect-1;
		n->cskip=p->cskip;
		n->creverse=p->creverse;
		p->suspended=NULL;
		p->next=n;
		return;
	}
	assert(p->creverse>0);
	//time to reverse rest of the stack
	while(p->cskip>0)
	{
		cresume_ops++;
		assert(p->suspended);
		pearl_resume(p->suspended);
		p->suspended=p->suspended->next;
		assert(p->suspended);
		p->cskip--;
	}
	//p->suspended is now last element i should generate
	tail=NULL;
	stack=p->suspended;
	cnt=p->creverse;
	while(cnt>0)
	{
		cresume_ops++;
		struct PEARL *n=new_pearl();
		assert(stack);
		pearl_resume(stack);
		n->next=tail;
		n->color=stack->color;
		tail=n;
		stack=stack->next;
		cnt--;
	}
	p->next=tail;
	p->suspended=NULL;
	cresume_ops++;
}

//create new necklace by deleting left- or rightmost pearl
struct NECKLACE* necklace_delete(struct NECKLACE *on,int left)
{
	struct NECKLACE *nn=new_necklace();
	nn->llen=on->llen;
	nn->rlen=on->rlen;
	nn->l=on->l;
	nn->r=on->r;
	assert(nn->rlen+nn->llen>=1);
	if(left)
	{
		//if left stack is empty, cut the right one
		if(nn->llen==0)
		{
			assert(nn->rlen>0);
			nn->rlen--;
		}
		//pop left stack
		else
		{
			pearl_resume(nn->l);
			nn->llen--;
			nn->l=nn->l->next;
		}
	}
	else
	{
		//if right stack is empty, cut the left one
		if(nn->rlen==0)
		{
			assert(nn->llen>0);
			nn->llen--;
		}
		//pop right stack
		else
		{
			pearl_resume(nn->r);
			nn->rlen--;
			nn->r=nn->r->next;
		}
	}
	necklace_rebalance(nn);
	cnecklace_removals++;
	return nn;
}

//create new necklace by appending single pearl to old necklace
struct NECKLACE *necklace_append(struct NECKLACE *on, int left, unsigned color)
{
	struct NECKLACE *nn=new_necklace();
	struct PEARL *p=new_pearl();
	nn->llen=on->llen;
	nn->rlen=on->rlen;
	nn->r=on->r;
	nn->l=on->l;
	p->color=color;
	if(left)
	{
		p->next=nn->l;
		nn->l=p;
		nn->llen++;
	}
	else
	{
		p->next=nn->r;
		nn->r=p;
		nn->rlen++;
	}
	necklace_rebalance(nn);
	cnecklace_appends++;
	return nn;
}

//initialize
void init()
{
	//init pearls
	cpearls=0;
	//init necklaces
	cnecklaces=1;
	necklaces[0].llen=0;
	necklaces[0].rlen=0;
	necklaces[0].l=NULL;
	necklaces[0].r=NULL;
	necklaces[0].num=0;
	//init debug counters
	cnecklace_appends=0;
	cnecklace_removals=0;
	cresume_ops=0;
}

//get left- or rightmost pearl color
int pearl(int neck_id, int on_left)
{
	struct NECKLACE *n=&(necklaces[neck_id]);
	assert(n->rlen+n->llen>0);
	if(on_left)
	{
		if(n->llen>0)
			return n->l->color;
		assert(n->rlen==1);
		return n->r->color;
	}
	else
	{
		if(n->rlen>0)
			return n->r->color;
		assert(n->llen==1);
		return n->l->color;
	}
}

//create new necklace from old one
void create (int from, char operation, int on_left, int param)
{
	assert(from<cnecklaces);
	assert((operation=='R')||(operation=='A'));
	if(operation=='R')
		necklace_delete(&(necklaces[from]),on_left);
	else
		necklace_append(&(necklaces[from]),on_left,param);
}

//recursive helper for printing right stack
int rec_append(struct PEARL *p,int count,char *strbuf,int strpos)
{
	int pos;
	if(count<=0)
		return strpos;
	assert(p!=NULL);
	pearl_resume(p);
	pos=rec_append(p->next,count-1,strbuf,strpos);
	pos+=sprintf(strbuf+pos,"%d ",p->color);
	return pos;
}

//debug routine -- returned string is global and reused by succesive necklace2str
const char* necklace2str(int neck_id)
{
	struct NECKLACE *n;
	struct PEARL *p;
	int i,strpos=0;
	assert(neck_id<cnecklaces);
	n=&(necklaces[neck_id]);
	if(n->llen+n->rlen>256000) 
	{
		sprintf(glob_str,"<%d pearls>",n->llen+n->rlen);
		return glob_str;
	}
	if(n->llen+n->rlen==0) 
	{
		sprintf(glob_str,"<empty>");
		return glob_str;
	}
	p=n->l;
	glob_str[strpos++]='[';
	glob_str[strpos++]=' ';
	for(i=n->llen;i>0;i--)
	{
		assert(p!=NULL);
		strpos+=sprintf(glob_str+strpos,"%d ",p->color);
		pearl_resume(p);
		p=p->next;
	}
	glob_str[strpos++]='|';
	glob_str[strpos++]=' ';
	strpos=rec_append(n->r,n->rlen,glob_str,strpos);
	glob_str[strpos-1]=' ';
	glob_str[strpos]=']';
	glob_str[strpos+1]=0;
	return glob_str;
}

//debug routine -- returned string is global and reused by succesive necklace2str
const char *necklace_status()
{
	sprintf(glob_str,"MEM: %d neckls used, %d pearls used, TIME: %d appends, %d rems, %d resume ops",
			cnecklaces,cpearls,cnecklace_appends,cnecklace_removals,cresume_ops);
	return glob_str;
}

